# This file is dynamically-generated; do not edit this file.
# TCL fragment to set some paths.

set_param board.repoPaths [list {board_files}];
set scriptdir {work_dir};
set wrkdir [file join $scriptdir obj]
set dcpdir {dcp_macro_dir};
set srcmainverilogfile "{verilog_files}";
set srcmainverilogfiles "{verilog_files}";
set top {top_module}
